"""Numba extension support for datetime.timedelta objects and their arrays.
"""
import datetime
import operator
from collections import namedtuple
import llvmlite.binding as ll
import numba
import numpy as np
import pandas as pd
from llvmlite import ir as lir
from numba.core import cgutils, types
from numba.core.imputils import lower_constant
from numba.extending import NativeValue, box, intrinsic, make_attribute_wrapper, models, overload, overload_attribute, overload_method, register_jitable, register_model, typeof_impl, unbox
from numba.parfors.array_analysis import ArrayAnalysis
import bodo
from bodo.hiframes.datetime_datetime_ext import datetime_datetime_type
from bodo.libs import hdatetime_ext
from bodo.utils.indexing import get_new_null_mask_bool_index, get_new_null_mask_int_index, get_new_null_mask_slice_index, setitem_slice_index_null_bits
from bodo.utils.typing import BodoError, get_overload_const_str, is_iterable_type, is_list_like_index_type, is_overload_constant_str
ll.add_symbol('box_datetime_timedelta_array', hdatetime_ext.
    box_datetime_timedelta_array)
ll.add_symbol('unbox_datetime_timedelta_array', hdatetime_ext.
    unbox_datetime_timedelta_array)


class NoInput:
    pass


_no_input = NoInput()


class NoInputType(types.Type):

    def __init__(self):
        super(NoInputType, self).__init__(name='NoInput')


register_model(NoInputType)(models.OpaqueModel)


@typeof_impl.register(NoInput)
def _typ_no_input(val, c):
    return NoInputType()


@lower_constant(NoInputType)
def constant_no_input(context, builder, ty, pyval):
    return context.get_dummy_value()


class PDTimeDeltaType(types.Type):

    def __init__(self):
        super(PDTimeDeltaType, self).__init__(name='PDTimeDeltaType()')


pd_timedelta_type = PDTimeDeltaType()
types.pd_timedelta_type = pd_timedelta_type


@typeof_impl.register(pd.Timedelta)
def typeof_pd_timedelta(val, c):
    return pd_timedelta_type


@register_model(PDTimeDeltaType)
class PDTimeDeltaModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        ciqsw__rdpa = [('value', types.int64)]
        super(PDTimeDeltaModel, self).__init__(dmm, fe_type, ciqsw__rdpa)


@box(PDTimeDeltaType)
def box_pd_timedelta(typ, val, c):
    hhh__zbadt = cgutils.create_struct_proxy(typ)(c.context, c.builder,
        value=val)
    qkjrd__dvfj = c.pyapi.long_from_longlong(hhh__zbadt.value)
    udnpb__yqyd = c.pyapi.unserialize(c.pyapi.serialize_object(pd.Timedelta))
    res = c.pyapi.call_function_objargs(udnpb__yqyd, (qkjrd__dvfj,))
    c.pyapi.decref(qkjrd__dvfj)
    c.pyapi.decref(udnpb__yqyd)
    return res


@unbox(PDTimeDeltaType)
def unbox_pd_timedelta(typ, val, c):
    qkjrd__dvfj = c.pyapi.object_getattr_string(val, 'value')
    clcu__cxlfu = c.pyapi.long_as_longlong(qkjrd__dvfj)
    hhh__zbadt = cgutils.create_struct_proxy(typ)(c.context, c.builder)
    hhh__zbadt.value = clcu__cxlfu
    c.pyapi.decref(qkjrd__dvfj)
    svnfd__ryt = cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return NativeValue(hhh__zbadt._getvalue(), is_error=svnfd__ryt)


@lower_constant(PDTimeDeltaType)
def lower_constant_pd_timedelta(context, builder, ty, pyval):
    value = context.get_constant(types.int64, pyval.value)
    return lir.Constant.literal_struct([value])


@overload(pd.Timedelta, no_unliteral=True)
def pd_timedelta(value=_no_input, unit='ns', days=0, seconds=0,
    microseconds=0, milliseconds=0, minutes=0, hours=0, weeks=0):
    if value == _no_input:

        def impl_timedelta_kw(value=_no_input, unit='ns', days=0, seconds=0,
            microseconds=0, milliseconds=0, minutes=0, hours=0, weeks=0):
            days += weeks * 7
            hours += days * 24
            minutes += 60 * hours
            seconds += 60 * minutes
            milliseconds += 1000 * seconds
            microseconds += 1000 * milliseconds
            kspu__fsxts = 1000 * microseconds
            return init_pd_timedelta(kspu__fsxts)
        return impl_timedelta_kw
    if value == bodo.string_type or is_overload_constant_str(value):

        def impl_str(value=_no_input, unit='ns', days=0, seconds=0,
            microseconds=0, milliseconds=0, minutes=0, hours=0, weeks=0):
            with numba.objmode(res='pd_timedelta_type'):
                res = pd.Timedelta(value)
            return res
        return impl_str
    if value == pd_timedelta_type:
        return (lambda value=_no_input, unit='ns', days=0, seconds=0,
            microseconds=0, milliseconds=0, minutes=0, hours=0, weeks=0: value)
    if value == datetime_timedelta_type:

        def impl_timedelta_datetime(value=_no_input, unit='ns', days=0,
            seconds=0, microseconds=0, milliseconds=0, minutes=0, hours=0,
            weeks=0):
            days = value.days
            seconds = 60 * 60 * 24 * days + value.seconds
            microseconds = 1000 * 1000 * seconds + value.microseconds
            kspu__fsxts = 1000 * microseconds
            return init_pd_timedelta(kspu__fsxts)
        return impl_timedelta_datetime
    if not is_overload_constant_str(unit):
        raise BodoError('pd.to_timedelta(): unit should be a constant string')
    unit = pd._libs.tslibs.timedeltas.parse_timedelta_unit(
        get_overload_const_str(unit))
    uoh__twj, tms__swa = pd._libs.tslibs.conversion.precision_from_unit(unit)

    def impl_timedelta(value=_no_input, unit='ns', days=0, seconds=0,
        microseconds=0, milliseconds=0, minutes=0, hours=0, weeks=0):
        return init_pd_timedelta(value * uoh__twj)
    return impl_timedelta


@intrinsic
def init_pd_timedelta(typingctx, value):

    def codegen(context, builder, signature, args):
        typ = signature.return_type
        timedelta = cgutils.create_struct_proxy(typ)(context, builder)
        timedelta.value = args[0]
        return timedelta._getvalue()
    return PDTimeDeltaType()(value), codegen


make_attribute_wrapper(PDTimeDeltaType, 'value', '_value')


@overload_attribute(PDTimeDeltaType, 'value')
@overload_attribute(PDTimeDeltaType, 'delta')
def pd_timedelta_get_value(td):

    def impl(td):
        return td._value
    return impl


@overload_attribute(PDTimeDeltaType, 'days')
def pd_timedelta_get_days(td):

    def impl(td):
        return td._value // (1000 * 1000 * 1000 * 60 * 60 * 24)
    return impl


@overload_attribute(PDTimeDeltaType, 'seconds')
def pd_timedelta_get_seconds(td):

    def impl(td):
        return td._value // (1000 * 1000 * 1000) % (60 * 60 * 24)
    return impl


@overload_attribute(PDTimeDeltaType, 'microseconds')
def pd_timedelta_get_microseconds(td):

    def impl(td):
        return td._value // 1000 % 1000000
    return impl


@overload_attribute(PDTimeDeltaType, 'nanoseconds')
def pd_timedelta_get_nanoseconds(td):

    def impl(td):
        return td._value % 1000
    return impl


@register_jitable
def _to_hours_pd_td(td):
    return td._value // (1000 * 1000 * 1000 * 60 * 60) % 24


@register_jitable
def _to_minutes_pd_td(td):
    return td._value // (1000 * 1000 * 1000 * 60) % 60


@register_jitable
def _to_seconds_pd_td(td):
    return td._value // (1000 * 1000 * 1000) % 60


@register_jitable
def _to_milliseconds_pd_td(td):
    return td._value // (1000 * 1000) % 1000


@register_jitable
def _to_microseconds_pd_td(td):
    return td._value // 1000 % 1000


Components = namedtuple('Components', ['days', 'hours', 'minutes',
    'seconds', 'milliseconds', 'microseconds', 'nanoseconds'], defaults=[0,
    0, 0, 0, 0, 0, 0])


@overload_attribute(PDTimeDeltaType, 'components', no_unliteral=True)
def pd_timedelta_get_components(td):

    def impl(td):
        a = Components(td.days, _to_hours_pd_td(td), _to_minutes_pd_td(td),
            _to_seconds_pd_td(td), _to_milliseconds_pd_td(td),
            _to_microseconds_pd_td(td), td.nanoseconds)
        return a
    return impl


@overload_method(PDTimeDeltaType, '__hash__', no_unliteral=True)
def pd_td___hash__(td):

    def impl(td):
        return hash(td._value)
    return impl


@overload_method(PDTimeDeltaType, 'to_numpy', no_unliteral=True)
@overload_method(PDTimeDeltaType, 'to_timedelta64', no_unliteral=True)
def pd_td_to_numpy(td):
    from bodo.hiframes.pd_timestamp_ext import integer_to_timedelta64

    def impl(td):
        return integer_to_timedelta64(td.value)
    return impl


@overload_method(PDTimeDeltaType, 'to_pytimedelta', no_unliteral=True)
def pd_td_to_pytimedelta(td):

    def impl(td):
        return datetime.timedelta(microseconds=np.int64(td._value / 1000))
    return impl


@overload_method(PDTimeDeltaType, 'total_seconds', no_unliteral=True)
def pd_td_total_seconds(td):

    def impl(td):
        return td._value // 1000 / 10 ** 6
    return impl


def overload_add_operator_datetime_timedelta(lhs, rhs):
    if lhs == pd_timedelta_type and rhs == pd_timedelta_type:

        def impl(lhs, rhs):
            val = lhs.value + rhs.value
            return pd.Timedelta(val)
        return impl
    if lhs == pd_timedelta_type and rhs == datetime_timedelta_type:

        def impl(lhs, rhs):
            pbh__xcxr = (rhs.microseconds + (rhs.seconds + rhs.days * 60 * 
                60 * 24) * 1000 * 1000) * 1000
            val = lhs.value + pbh__xcxr
            return pd.Timedelta(val)
        return impl
    if lhs == datetime_timedelta_type and rhs == pd_timedelta_type:

        def impl(lhs, rhs):
            vwszm__fjjp = (lhs.microseconds + (lhs.seconds + lhs.days * 60 *
                60 * 24) * 1000 * 1000) * 1000
            val = vwszm__fjjp + rhs.value
            return pd.Timedelta(val)
        return impl
    if lhs == pd_timedelta_type and rhs == datetime_datetime_type:
        from bodo.hiframes.pd_timestamp_ext import compute_pd_timestamp

        def impl(lhs, rhs):
            wski__covmf = rhs.toordinal()
            dgglz__mrbop = rhs.second + rhs.minute * 60 + rhs.hour * 3600
            dvlff__wpkf = rhs.microsecond
            rwgm__cuy = lhs.value // 1000
            hdx__jtzn = lhs.nanoseconds
            rsgeh__eextq = dvlff__wpkf + rwgm__cuy
            xfzaa__dfqrh = 1000000 * (wski__covmf * 86400 + dgglz__mrbop
                ) + rsgeh__eextq
            pllvv__hsj = hdx__jtzn
            return compute_pd_timestamp(xfzaa__dfqrh, pllvv__hsj)
        return impl
    if lhs == datetime_datetime_type and rhs == pd_timedelta_type:

        def impl(lhs, rhs):
            return lhs + rhs.to_pytimedelta()
        return impl
    if lhs == datetime_timedelta_type and rhs == datetime_timedelta_type:

        def impl(lhs, rhs):
            d = lhs.days + rhs.days
            s = lhs.seconds + rhs.seconds
            us = lhs.microseconds + rhs.microseconds
            return datetime.timedelta(d, s, us)
        return impl
    if lhs == datetime_timedelta_type and rhs == datetime_datetime_type:

        def impl(lhs, rhs):
            hmw__lmjdt = datetime.timedelta(rhs.toordinal(), hours=rhs.hour,
                minutes=rhs.minute, seconds=rhs.second, microseconds=rhs.
                microsecond)
            hmw__lmjdt = hmw__lmjdt + lhs
            org__cjf, uxdf__mdxpl = divmod(hmw__lmjdt.seconds, 3600)
            nduyn__ugoe, mruag__syrh = divmod(uxdf__mdxpl, 60)
            if 0 < hmw__lmjdt.days <= _MAXORDINAL:
                d = bodo.hiframes.datetime_date_ext.fromordinal_impl(hmw__lmjdt
                    .days)
                return datetime.datetime(d.year, d.month, d.day, org__cjf,
                    nduyn__ugoe, mruag__syrh, hmw__lmjdt.microseconds)
            raise OverflowError('result out of range')
        return impl
    if lhs == datetime_datetime_type and rhs == datetime_timedelta_type:

        def impl(lhs, rhs):
            hmw__lmjdt = datetime.timedelta(lhs.toordinal(), hours=lhs.hour,
                minutes=lhs.minute, seconds=lhs.second, microseconds=lhs.
                microsecond)
            hmw__lmjdt = hmw__lmjdt + rhs
            org__cjf, uxdf__mdxpl = divmod(hmw__lmjdt.seconds, 3600)
            nduyn__ugoe, mruag__syrh = divmod(uxdf__mdxpl, 60)
            if 0 < hmw__lmjdt.days <= _MAXORDINAL:
                d = bodo.hiframes.datetime_date_ext.fromordinal_impl(hmw__lmjdt
                    .days)
                return datetime.datetime(d.year, d.month, d.day, org__cjf,
                    nduyn__ugoe, mruag__syrh, hmw__lmjdt.microseconds)
            raise OverflowError('result out of range')
        return impl


def overload_sub_operator_datetime_timedelta(lhs, rhs):
    if lhs == pd_timedelta_type and rhs == pd_timedelta_type:

        def impl(lhs, rhs):
            wwlkq__acqje = lhs.value - rhs.value
            return pd.Timedelta(wwlkq__acqje)
        return impl
    if lhs == pd_timedelta_type and rhs == datetime_timedelta_type:

        def impl(lhs, rhs):
            return lhs + -rhs
        return impl
    if lhs == datetime_timedelta_type and rhs == pd_timedelta_type:

        def impl(lhs, rhs):
            return lhs + -rhs
        return impl
    if lhs == datetime_datetime_type and rhs == pd_timedelta_type:

        def impl(lhs, rhs):
            return lhs + -rhs
        return impl
    if lhs == datetime_timedelta_type and rhs == datetime_timedelta_type:

        def impl(lhs, rhs):
            d = lhs.days - rhs.days
            s = lhs.seconds - rhs.seconds
            us = lhs.microseconds - rhs.microseconds
            return datetime.timedelta(d, s, us)
        return impl
    if lhs == datetime_datetime_type and rhs == datetime_timedelta_type:

        def impl(lhs, rhs):
            return lhs + -rhs
        return impl
    if lhs == datetime_timedelta_array_type and rhs == datetime_timedelta_type:

        def impl(lhs, rhs):
            bht__pss = lhs
            numba.parfors.parfor.init_prange()
            n = len(bht__pss)
            A = alloc_datetime_timedelta_array(n)
            for uvhh__kei in numba.parfors.parfor.internal_prange(n):
                A[uvhh__kei] = bht__pss[uvhh__kei] - rhs
            return A
        return impl


def overload_mul_operator_timedelta(lhs, rhs):
    if lhs == pd_timedelta_type and isinstance(rhs, types.Integer):

        def impl(lhs, rhs):
            return pd.Timedelta(lhs.value * rhs)
        return impl
    elif isinstance(lhs, types.Integer) and rhs == pd_timedelta_type:

        def impl(lhs, rhs):
            return pd.Timedelta(rhs.value * lhs)
        return impl
    if lhs == datetime_timedelta_type and isinstance(rhs, types.Integer):

        def impl(lhs, rhs):
            d = lhs.days * rhs
            s = lhs.seconds * rhs
            us = lhs.microseconds * rhs
            return datetime.timedelta(d, s, us)
        return impl
    elif isinstance(lhs, types.Integer) and rhs == datetime_timedelta_type:

        def impl(lhs, rhs):
            d = lhs * rhs.days
            s = lhs * rhs.seconds
            us = lhs * rhs.microseconds
            return datetime.timedelta(d, s, us)
        return impl


def overload_floordiv_operator_pd_timedelta(lhs, rhs):
    if lhs == pd_timedelta_type and rhs == pd_timedelta_type:

        def impl(lhs, rhs):
            return lhs.value // rhs.value
        return impl
    elif lhs == pd_timedelta_type and isinstance(rhs, types.Integer):

        def impl(lhs, rhs):
            return pd.Timedelta(lhs.value // rhs)
        return impl


def overload_truediv_operator_pd_timedelta(lhs, rhs):
    if lhs == pd_timedelta_type and rhs == pd_timedelta_type:

        def impl(lhs, rhs):
            return lhs.value / rhs.value
        return impl
    elif lhs == pd_timedelta_type and isinstance(rhs, types.Integer):

        def impl(lhs, rhs):
            return pd.Timedelta(int(lhs.value / rhs))
        return impl


def overload_mod_operator_timedeltas(lhs, rhs):
    if lhs == pd_timedelta_type and rhs == pd_timedelta_type:

        def impl(lhs, rhs):
            return pd.Timedelta(lhs.value % rhs.value)
        return impl
    if lhs == datetime_timedelta_type and rhs == datetime_timedelta_type:

        def impl(lhs, rhs):
            csj__efdda = _to_microseconds(lhs) % _to_microseconds(rhs)
            return datetime.timedelta(0, 0, csj__efdda)
        return impl


def pd_create_cmp_op_overload(op):

    def overload_pd_timedelta_cmp(lhs, rhs):
        if lhs == pd_timedelta_type and rhs == pd_timedelta_type:

            def impl(lhs, rhs):
                return op(lhs.value, rhs.value)
            return impl
        if lhs == pd_timedelta_type and rhs == bodo.timedelta64ns:
            return lambda lhs, rhs: op(bodo.hiframes.pd_timestamp_ext.
                integer_to_timedelta64(lhs.value), rhs)
        if lhs == bodo.timedelta64ns and rhs == pd_timedelta_type:
            return lambda lhs, rhs: op(lhs, bodo.hiframes.pd_timestamp_ext.
                integer_to_timedelta64(rhs.value))
    return overload_pd_timedelta_cmp


@overload(operator.neg, no_unliteral=True)
def pd_timedelta_neg(lhs):
    if lhs == pd_timedelta_type:

        def impl(lhs):
            return pd.Timedelta(-lhs.value)
        return impl


@overload(operator.pos, no_unliteral=True)
def pd_timedelta_pos(lhs):
    if lhs == pd_timedelta_type:

        def impl(lhs):
            return lhs
        return impl


@overload(divmod, no_unliteral=True)
def pd_timedelta_divmod(lhs, rhs):
    if lhs == pd_timedelta_type and rhs == pd_timedelta_type:

        def impl(lhs, rhs):
            nnmbo__kjrnn, csj__efdda = divmod(lhs.value, rhs.value)
            return nnmbo__kjrnn, pd.Timedelta(csj__efdda)
        return impl


@overload(abs, no_unliteral=True)
def pd_timedelta_abs(lhs):
    if lhs == pd_timedelta_type:

        def impl(lhs):
            if lhs.value < 0:
                return -lhs
            else:
                return lhs
        return impl


class DatetimeTimeDeltaType(types.Type):

    def __init__(self):
        super(DatetimeTimeDeltaType, self).__init__(name=
            'DatetimeTimeDeltaType()')


datetime_timedelta_type = DatetimeTimeDeltaType()


@typeof_impl.register(datetime.timedelta)
def typeof_datetime_timedelta(val, c):
    return datetime_timedelta_type


@register_model(DatetimeTimeDeltaType)
class DatetimeTimeDeltaModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        ciqsw__rdpa = [('days', types.int64), ('seconds', types.int64), (
            'microseconds', types.int64)]
        super(DatetimeTimeDeltaModel, self).__init__(dmm, fe_type, ciqsw__rdpa)


@box(DatetimeTimeDeltaType)
def box_datetime_timedelta(typ, val, c):
    hhh__zbadt = cgutils.create_struct_proxy(typ)(c.context, c.builder,
        value=val)
    yrqb__vjzbf = c.pyapi.long_from_longlong(hhh__zbadt.days)
    xovkq__fxum = c.pyapi.long_from_longlong(hhh__zbadt.seconds)
    xcfa__kvkdo = c.pyapi.long_from_longlong(hhh__zbadt.microseconds)
    udnpb__yqyd = c.pyapi.unserialize(c.pyapi.serialize_object(datetime.
        timedelta))
    res = c.pyapi.call_function_objargs(udnpb__yqyd, (yrqb__vjzbf,
        xovkq__fxum, xcfa__kvkdo))
    c.pyapi.decref(yrqb__vjzbf)
    c.pyapi.decref(xovkq__fxum)
    c.pyapi.decref(xcfa__kvkdo)
    c.pyapi.decref(udnpb__yqyd)
    return res


@unbox(DatetimeTimeDeltaType)
def unbox_datetime_timedelta(typ, val, c):
    yrqb__vjzbf = c.pyapi.object_getattr_string(val, 'days')
    xovkq__fxum = c.pyapi.object_getattr_string(val, 'seconds')
    xcfa__kvkdo = c.pyapi.object_getattr_string(val, 'microseconds')
    slqrt__jgpyy = c.pyapi.long_as_longlong(yrqb__vjzbf)
    zylz__naet = c.pyapi.long_as_longlong(xovkq__fxum)
    rhkf__exuh = c.pyapi.long_as_longlong(xcfa__kvkdo)
    hhh__zbadt = cgutils.create_struct_proxy(typ)(c.context, c.builder)
    hhh__zbadt.days = slqrt__jgpyy
    hhh__zbadt.seconds = zylz__naet
    hhh__zbadt.microseconds = rhkf__exuh
    c.pyapi.decref(yrqb__vjzbf)
    c.pyapi.decref(xovkq__fxum)
    c.pyapi.decref(xcfa__kvkdo)
    svnfd__ryt = cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return NativeValue(hhh__zbadt._getvalue(), is_error=svnfd__ryt)


@lower_constant(DatetimeTimeDeltaType)
def lower_constant_datetime_timedelta(context, builder, ty, pyval):
    days = context.get_constant(types.int64, pyval.days)
    seconds = context.get_constant(types.int64, pyval.seconds)
    microseconds = context.get_constant(types.int64, pyval.microseconds)
    return lir.Constant.literal_struct([days, seconds, microseconds])


@overload(datetime.timedelta, no_unliteral=True)
def datetime_timedelta(days=0, seconds=0, microseconds=0, milliseconds=0,
    minutes=0, hours=0, weeks=0):

    def impl_timedelta(days=0, seconds=0, microseconds=0, milliseconds=0,
        minutes=0, hours=0, weeks=0):
        d = s = us = 0
        days += weeks * 7
        seconds += minutes * 60 + hours * 3600
        microseconds += milliseconds * 1000
        d = days
        days, seconds = divmod(seconds, 24 * 3600)
        d += days
        s += int(seconds)
        seconds, us = divmod(microseconds, 1000000)
        days, seconds = divmod(seconds, 24 * 3600)
        d += days
        s += seconds
        return init_timedelta(d, s, us)
    return impl_timedelta


@intrinsic
def init_timedelta(typingctx, d, s, us):

    def codegen(context, builder, signature, args):
        typ = signature.return_type
        timedelta = cgutils.create_struct_proxy(typ)(context, builder)
        timedelta.days = args[0]
        timedelta.seconds = args[1]
        timedelta.microseconds = args[2]
        return timedelta._getvalue()
    return DatetimeTimeDeltaType()(d, s, us), codegen


make_attribute_wrapper(DatetimeTimeDeltaType, 'days', '_days')
make_attribute_wrapper(DatetimeTimeDeltaType, 'seconds', '_seconds')
make_attribute_wrapper(DatetimeTimeDeltaType, 'microseconds', '_microseconds')


@overload_attribute(DatetimeTimeDeltaType, 'days')
def timedelta_get_days(td):

    def impl(td):
        return td._days
    return impl


@overload_attribute(DatetimeTimeDeltaType, 'seconds')
def timedelta_get_seconds(td):

    def impl(td):
        return td._seconds
    return impl


@overload_attribute(DatetimeTimeDeltaType, 'microseconds')
def timedelta_get_microseconds(td):

    def impl(td):
        return td._microseconds
    return impl


@overload_method(DatetimeTimeDeltaType, 'total_seconds', no_unliteral=True)
def total_seconds(td):

    def impl(td):
        return ((td._days * 86400 + td._seconds) * 10 ** 6 + td._microseconds
            ) / 10 ** 6
    return impl


@overload_method(DatetimeTimeDeltaType, '__hash__', no_unliteral=True)
def __hash__(td):

    def impl(td):
        return hash((td._days, td._seconds, td._microseconds))
    return impl


@register_jitable
def _to_nanoseconds(td):
    return np.int64(((td._days * 86400 + td._seconds) * 1000000 + td.
        _microseconds) * 1000)


@register_jitable
def _to_microseconds(td):
    return (td._days * (24 * 3600) + td._seconds) * 1000000 + td._microseconds


@register_jitable
def _cmp(x, y):
    return 0 if x == y else 1 if x > y else -1


@register_jitable
def _getstate(td):
    return td._days, td._seconds, td._microseconds


@register_jitable
def _divide_and_round(a, b):
    nnmbo__kjrnn, csj__efdda = divmod(a, b)
    csj__efdda *= 2
    kksa__trb = csj__efdda > b if b > 0 else csj__efdda < b
    if kksa__trb or csj__efdda == b and nnmbo__kjrnn % 2 == 1:
        nnmbo__kjrnn += 1
    return nnmbo__kjrnn


_MAXORDINAL = 3652059


def overload_floordiv_operator_dt_timedelta(lhs, rhs):
    if lhs == datetime_timedelta_type and rhs == datetime_timedelta_type:

        def impl(lhs, rhs):
            us = _to_microseconds(lhs)
            return us // _to_microseconds(rhs)
        return impl
    elif lhs == datetime_timedelta_type and rhs == types.int64:

        def impl(lhs, rhs):
            us = _to_microseconds(lhs)
            return datetime.timedelta(0, 0, us // rhs)
        return impl


def overload_truediv_operator_dt_timedelta(lhs, rhs):
    if lhs == datetime_timedelta_type and rhs == datetime_timedelta_type:

        def impl(lhs, rhs):
            us = _to_microseconds(lhs)
            return us / _to_microseconds(rhs)
        return impl
    elif lhs == datetime_timedelta_type and rhs == types.int64:

        def impl(lhs, rhs):
            us = _to_microseconds(lhs)
            return datetime.timedelta(0, 0, _divide_and_round(us, rhs))
        return impl


def create_cmp_op_overload(op):

    def overload_timedelta_cmp(lhs, rhs):
        if lhs == datetime_timedelta_type and rhs == datetime_timedelta_type:

            def impl(lhs, rhs):
                omy__asob = _cmp(_getstate(lhs), _getstate(rhs))
                return op(omy__asob, 0)
            return impl
    return overload_timedelta_cmp


@overload(operator.neg, no_unliteral=True)
def timedelta_neg(lhs):
    if lhs == datetime_timedelta_type:

        def impl(lhs):
            return datetime.timedelta(-lhs.days, -lhs.seconds, -lhs.
                microseconds)
        return impl


@overload(operator.pos, no_unliteral=True)
def timedelta_pos(lhs):
    if lhs == datetime_timedelta_type:

        def impl(lhs):
            return lhs
        return impl


@overload(divmod, no_unliteral=True)
def timedelta_divmod(lhs, rhs):
    if lhs == datetime_timedelta_type and rhs == datetime_timedelta_type:

        def impl(lhs, rhs):
            nnmbo__kjrnn, csj__efdda = divmod(_to_microseconds(lhs),
                _to_microseconds(rhs))
            return nnmbo__kjrnn, datetime.timedelta(0, 0, csj__efdda)
        return impl


@overload(abs, no_unliteral=True)
def timedelta_abs(lhs):
    if lhs == datetime_timedelta_type:

        def impl(lhs):
            if lhs.days < 0:
                return -lhs
            else:
                return lhs
        return impl


@intrinsic
def cast_numpy_timedelta_to_int(typingctx, val=None):
    assert val in (types.NPTimedelta('ns'), types.int64)

    def codegen(context, builder, signature, args):
        return args[0]
    return types.int64(val), codegen


@overload(bool, no_unliteral=True)
def timedelta_to_bool(timedelta):
    if timedelta != datetime_timedelta_type:
        return
    hyyrt__tyges = datetime.timedelta(0)

    def impl(timedelta):
        return timedelta != hyyrt__tyges
    return impl


class DatetimeTimeDeltaArrayType(types.ArrayCompatible):

    def __init__(self):
        super(DatetimeTimeDeltaArrayType, self).__init__(name=
            'DatetimeTimeDeltaArrayType()')

    @property
    def as_array(self):
        return types.Array(types.undefined, 1, 'C')

    @property
    def dtype(self):
        return datetime_timedelta_type

    def copy(self):
        return DatetimeTimeDeltaArrayType()


datetime_timedelta_array_type = DatetimeTimeDeltaArrayType()
types.datetime_timedelta_array_type = datetime_timedelta_array_type
days_data_type = types.Array(types.int64, 1, 'C')
seconds_data_type = types.Array(types.int64, 1, 'C')
microseconds_data_type = types.Array(types.int64, 1, 'C')
nulls_type = types.Array(types.uint8, 1, 'C')


@register_model(DatetimeTimeDeltaArrayType)
class DatetimeTimeDeltaArrayModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        ciqsw__rdpa = [('days_data', days_data_type), ('seconds_data',
            seconds_data_type), ('microseconds_data',
            microseconds_data_type), ('null_bitmap', nulls_type)]
        models.StructModel.__init__(self, dmm, fe_type, ciqsw__rdpa)


make_attribute_wrapper(DatetimeTimeDeltaArrayType, 'days_data', '_days_data')
make_attribute_wrapper(DatetimeTimeDeltaArrayType, 'seconds_data',
    '_seconds_data')
make_attribute_wrapper(DatetimeTimeDeltaArrayType, 'microseconds_data',
    '_microseconds_data')
make_attribute_wrapper(DatetimeTimeDeltaArrayType, 'null_bitmap',
    '_null_bitmap')


@overload_method(DatetimeTimeDeltaArrayType, 'copy', no_unliteral=True)
def overload_datetime_timedelta_arr_copy(A):
    return (lambda A: bodo.hiframes.datetime_timedelta_ext.
        init_datetime_timedelta_array(A._days_data.copy(), A._seconds_data.
        copy(), A._microseconds_data.copy(), A._null_bitmap.copy()))


@unbox(DatetimeTimeDeltaArrayType)
def unbox_datetime_timedelta_array(typ, val, c):
    n = bodo.utils.utils.object_length(c, val)
    hxuq__meg = types.Array(types.intp, 1, 'C')
    vnaz__daz = bodo.utils.utils._empty_nd_impl(c.context, c.builder,
        hxuq__meg, [n])
    wmda__mmtuh = bodo.utils.utils._empty_nd_impl(c.context, c.builder,
        hxuq__meg, [n])
    xjya__sgzma = bodo.utils.utils._empty_nd_impl(c.context, c.builder,
        hxuq__meg, [n])
    odwqz__xdlm = c.builder.udiv(c.builder.add(n, lir.Constant(lir.IntType(
        64), 7)), lir.Constant(lir.IntType(64), 8))
    llcqz__tlytv = bodo.utils.utils._empty_nd_impl(c.context, c.builder,
        types.Array(types.uint8, 1, 'C'), [odwqz__xdlm])
    xaf__swsz = lir.FunctionType(lir.VoidType(), [lir.IntType(8).as_pointer
        (), lir.IntType(64), lir.IntType(64).as_pointer(), lir.IntType(64).
        as_pointer(), lir.IntType(64).as_pointer(), lir.IntType(8).
        as_pointer()])
    qgk__qkfx = cgutils.get_or_insert_function(c.builder.module, xaf__swsz,
        name='unbox_datetime_timedelta_array')
    c.builder.call(qgk__qkfx, [val, n, vnaz__daz.data, wmda__mmtuh.data,
        xjya__sgzma.data, llcqz__tlytv.data])
    hhvvv__pwlrs = cgutils.create_struct_proxy(typ)(c.context, c.builder)
    hhvvv__pwlrs.days_data = vnaz__daz._getvalue()
    hhvvv__pwlrs.seconds_data = wmda__mmtuh._getvalue()
    hhvvv__pwlrs.microseconds_data = xjya__sgzma._getvalue()
    hhvvv__pwlrs.null_bitmap = llcqz__tlytv._getvalue()
    svnfd__ryt = cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return NativeValue(hhvvv__pwlrs._getvalue(), is_error=svnfd__ryt)


@box(DatetimeTimeDeltaArrayType)
def box_datetime_timedelta_array(typ, val, c):
    bht__pss = cgutils.create_struct_proxy(typ)(c.context, c.builder, val)
    vnaz__daz = c.context.make_array(types.Array(types.int64, 1, 'C'))(c.
        context, c.builder, bht__pss.days_data)
    wmda__mmtuh = c.context.make_array(types.Array(types.int64, 1, 'C'))(c.
        context, c.builder, bht__pss.seconds_data).data
    xjya__sgzma = c.context.make_array(types.Array(types.int64, 1, 'C'))(c.
        context, c.builder, bht__pss.microseconds_data).data
    qxce__zmku = c.context.make_array(types.Array(types.uint8, 1, 'C'))(c.
        context, c.builder, bht__pss.null_bitmap).data
    n = c.builder.extract_value(vnaz__daz.shape, 0)
    xaf__swsz = lir.FunctionType(c.pyapi.pyobj, [lir.IntType(64), lir.
        IntType(64).as_pointer(), lir.IntType(64).as_pointer(), lir.IntType
        (64).as_pointer(), lir.IntType(8).as_pointer()])
    dont__hle = cgutils.get_or_insert_function(c.builder.module, xaf__swsz,
        name='box_datetime_timedelta_array')
    mdo__wby = c.builder.call(dont__hle, [n, vnaz__daz.data, wmda__mmtuh,
        xjya__sgzma, qxce__zmku])
    c.context.nrt.decref(c.builder, typ, val)
    return mdo__wby


@intrinsic
def init_datetime_timedelta_array(typingctx, days_data, seconds_data,
    microseconds_data, nulls=None):
    assert days_data == types.Array(types.int64, 1, 'C')
    assert seconds_data == types.Array(types.int64, 1, 'C')
    assert microseconds_data == types.Array(types.int64, 1, 'C')
    assert nulls == types.Array(types.uint8, 1, 'C')

    def codegen(context, builder, signature, args):
        ktsuz__twseg, sqevd__nggx, jprg__ofrw, cwt__yiz = args
        qoh__kud = cgutils.create_struct_proxy(signature.return_type)(context,
            builder)
        qoh__kud.days_data = ktsuz__twseg
        qoh__kud.seconds_data = sqevd__nggx
        qoh__kud.microseconds_data = jprg__ofrw
        qoh__kud.null_bitmap = cwt__yiz
        context.nrt.incref(builder, signature.args[0], ktsuz__twseg)
        context.nrt.incref(builder, signature.args[1], sqevd__nggx)
        context.nrt.incref(builder, signature.args[2], jprg__ofrw)
        context.nrt.incref(builder, signature.args[3], cwt__yiz)
        return qoh__kud._getvalue()
    osf__apwa = datetime_timedelta_array_type(days_data, seconds_data,
        microseconds_data, nulls)
    return osf__apwa, codegen


@lower_constant(DatetimeTimeDeltaArrayType)
def lower_constant_datetime_timedelta_arr(context, builder, typ, pyval):
    n = len(pyval)
    vnaz__daz = np.empty(n, np.int64)
    wmda__mmtuh = np.empty(n, np.int64)
    xjya__sgzma = np.empty(n, np.int64)
    sin__nyt = np.empty(n + 7 >> 3, np.uint8)
    for uvhh__kei, s in enumerate(pyval):
        cdy__wzslm = pd.isna(s)
        bodo.libs.int_arr_ext.set_bit_to_arr(sin__nyt, uvhh__kei, int(not
            cdy__wzslm))
        if not cdy__wzslm:
            vnaz__daz[uvhh__kei] = s.days
            wmda__mmtuh[uvhh__kei] = s.seconds
            xjya__sgzma[uvhh__kei] = s.microseconds
    gzzgr__jbr = context.get_constant_generic(builder, days_data_type,
        vnaz__daz)
    yazh__ilfzr = context.get_constant_generic(builder, seconds_data_type,
        wmda__mmtuh)
    artmy__vxj = context.get_constant_generic(builder,
        microseconds_data_type, xjya__sgzma)
    dvzv__erwd = context.get_constant_generic(builder, nulls_type, sin__nyt)
    return lir.Constant.literal_struct([gzzgr__jbr, yazh__ilfzr, artmy__vxj,
        dvzv__erwd])


@numba.njit(no_cpython_wrapper=True)
def alloc_datetime_timedelta_array(n):
    vnaz__daz = np.empty(n, dtype=np.int64)
    wmda__mmtuh = np.empty(n, dtype=np.int64)
    xjya__sgzma = np.empty(n, dtype=np.int64)
    nulls = np.full(n + 7 >> 3, 255, np.uint8)
    return init_datetime_timedelta_array(vnaz__daz, wmda__mmtuh,
        xjya__sgzma, nulls)


def alloc_datetime_timedelta_array_equiv(self, scope, equiv_set, loc, args, kws
    ):
    assert len(args) == 1 and not kws
    return ArrayAnalysis.AnalyzeResult(shape=args[0], pre=[])


(ArrayAnalysis.
    _analyze_op_call_bodo_hiframes_datetime_timedelta_ext_alloc_datetime_timedelta_array
    ) = alloc_datetime_timedelta_array_equiv


@overload(operator.getitem, no_unliteral=True)
def dt_timedelta_arr_getitem(A, ind):
    if A != datetime_timedelta_array_type:
        return
    if isinstance(ind, types.Integer):

        def impl_int(A, ind):
            return datetime.timedelta(days=A._days_data[ind], seconds=A.
                _seconds_data[ind], microseconds=A._microseconds_data[ind])
        return impl_int
    if is_list_like_index_type(ind) and ind.dtype == types.bool_:

        def impl_bool(A, ind):
            syq__gwobk = bodo.utils.conversion.coerce_to_ndarray(ind)
            whfux__vqk = A._null_bitmap
            mdvxf__lrbcp = A._days_data[syq__gwobk]
            kjd__xdzn = A._seconds_data[syq__gwobk]
            hrd__xgilz = A._microseconds_data[syq__gwobk]
            n = len(mdvxf__lrbcp)
            udh__etttn = get_new_null_mask_bool_index(whfux__vqk, ind, n)
            return init_datetime_timedelta_array(mdvxf__lrbcp, kjd__xdzn,
                hrd__xgilz, udh__etttn)
        return impl_bool
    if is_list_like_index_type(ind) and isinstance(ind.dtype, types.Integer):

        def impl(A, ind):
            syq__gwobk = bodo.utils.conversion.coerce_to_ndarray(ind)
            whfux__vqk = A._null_bitmap
            mdvxf__lrbcp = A._days_data[syq__gwobk]
            kjd__xdzn = A._seconds_data[syq__gwobk]
            hrd__xgilz = A._microseconds_data[syq__gwobk]
            n = len(mdvxf__lrbcp)
            udh__etttn = get_new_null_mask_int_index(whfux__vqk, syq__gwobk, n)
            return init_datetime_timedelta_array(mdvxf__lrbcp, kjd__xdzn,
                hrd__xgilz, udh__etttn)
        return impl
    if isinstance(ind, types.SliceType):

        def impl_slice(A, ind):
            n = len(A._days_data)
            whfux__vqk = A._null_bitmap
            mdvxf__lrbcp = np.ascontiguousarray(A._days_data[ind])
            kjd__xdzn = np.ascontiguousarray(A._seconds_data[ind])
            hrd__xgilz = np.ascontiguousarray(A._microseconds_data[ind])
            udh__etttn = get_new_null_mask_slice_index(whfux__vqk, ind, n)
            return init_datetime_timedelta_array(mdvxf__lrbcp, kjd__xdzn,
                hrd__xgilz, udh__etttn)
        return impl_slice
    raise BodoError(
        f'getitem for DatetimeTimedeltaArray with indexing type {ind} not supported.'
        )


@overload(operator.setitem, no_unliteral=True)
def dt_timedelta_arr_setitem(A, ind, val):
    if A != datetime_timedelta_array_type:
        return
    if val == types.none or isinstance(val, types.optional):
        return
    urhn__ughx = (
        f"setitem for DatetimeTimedeltaArray with indexing type {ind} received an incorrect 'value' type {val}."
        )
    if isinstance(ind, types.Integer):
        if types.unliteral(val) == datetime_timedelta_type:

            def impl(A, ind, val):
                A._days_data[ind] = val._days
                A._seconds_data[ind] = val._seconds
                A._microseconds_data[ind] = val._microseconds
                bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap, ind, 1)
            return impl
        else:
            raise BodoError(urhn__ughx)
    if not (is_iterable_type(val) and val.dtype == bodo.
        datetime_timedelta_type or types.unliteral(val) ==
        datetime_timedelta_type):
        raise BodoError(urhn__ughx)
    if is_list_like_index_type(ind) and isinstance(ind.dtype, types.Integer):
        if types.unliteral(val) == datetime_timedelta_type:

            def impl_arr_ind_scalar(A, ind, val):
                n = len(A)
                for uvhh__kei in range(n):
                    A._days_data[ind[uvhh__kei]] = val._days
                    A._seconds_data[ind[uvhh__kei]] = val._seconds
                    A._microseconds_data[ind[uvhh__kei]] = val._microseconds
                    bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                        ind[uvhh__kei], 1)
            return impl_arr_ind_scalar
        else:

            def impl_arr_ind(A, ind, val):
                val = bodo.utils.conversion.coerce_to_array(val,
                    use_nullable_array=True)
                n = len(val._days_data)
                for uvhh__kei in range(n):
                    A._days_data[ind[uvhh__kei]] = val._days_data[uvhh__kei]
                    A._seconds_data[ind[uvhh__kei]] = val._seconds_data[
                        uvhh__kei]
                    A._microseconds_data[ind[uvhh__kei]
                        ] = val._microseconds_data[uvhh__kei]
                    hwsy__seg = bodo.libs.int_arr_ext.get_bit_bitmap_arr(val
                        ._null_bitmap, uvhh__kei)
                    bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                        ind[uvhh__kei], hwsy__seg)
            return impl_arr_ind
    if is_list_like_index_type(ind) and ind.dtype == types.bool_:
        if types.unliteral(val) == datetime_timedelta_type:

            def impl_bool_ind_mask_scalar(A, ind, val):
                n = len(ind)
                for uvhh__kei in range(n):
                    if not bodo.libs.array_kernels.isna(ind, uvhh__kei
                        ) and ind[uvhh__kei]:
                        A._days_data[uvhh__kei] = val._days
                        A._seconds_data[uvhh__kei] = val._seconds
                        A._microseconds_data[uvhh__kei] = val._microseconds
                        bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                            uvhh__kei, 1)
            return impl_bool_ind_mask_scalar
        else:

            def impl_bool_ind_mask(A, ind, val):
                val = bodo.utils.conversion.coerce_to_array(val,
                    use_nullable_array=True)
                n = len(ind)
                illrl__jjypq = 0
                for uvhh__kei in range(n):
                    if not bodo.libs.array_kernels.isna(ind, uvhh__kei
                        ) and ind[uvhh__kei]:
                        A._days_data[uvhh__kei] = val._days_data[illrl__jjypq]
                        A._seconds_data[uvhh__kei] = val._seconds_data[
                            illrl__jjypq]
                        A._microseconds_data[uvhh__kei
                            ] = val._microseconds_data[illrl__jjypq]
                        hwsy__seg = bodo.libs.int_arr_ext.get_bit_bitmap_arr(
                            val._null_bitmap, illrl__jjypq)
                        bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                            uvhh__kei, hwsy__seg)
                        illrl__jjypq += 1
            return impl_bool_ind_mask
    if isinstance(ind, types.SliceType):
        if types.unliteral(val) == datetime_timedelta_type:

            def impl_slice_scalar(A, ind, val):
                mqav__mlly = numba.cpython.unicode._normalize_slice(ind, len(A)
                    )
                for uvhh__kei in range(mqav__mlly.start, mqav__mlly.stop,
                    mqav__mlly.step):
                    A._days_data[uvhh__kei] = val._days
                    A._seconds_data[uvhh__kei] = val._seconds
                    A._microseconds_data[uvhh__kei] = val._microseconds
                    bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                        uvhh__kei, 1)
            return impl_slice_scalar
        else:

            def impl_slice_mask(A, ind, val):
                val = bodo.utils.conversion.coerce_to_array(val,
                    use_nullable_array=True)
                n = len(A._days_data)
                A._days_data[ind] = val._days_data
                A._seconds_data[ind] = val._seconds_data
                A._microseconds_data[ind] = val._microseconds_data
                mqkfp__thw = val._null_bitmap.copy()
                setitem_slice_index_null_bits(A._null_bitmap, mqkfp__thw,
                    ind, n)
            return impl_slice_mask
    raise BodoError(
        f'setitem for DatetimeTimedeltaArray with indexing type {ind} not supported.'
        )


@overload(len, no_unliteral=True)
def overload_len_datetime_timedelta_arr(A):
    if A == datetime_timedelta_array_type:
        return lambda A: len(A._days_data)


@overload_attribute(DatetimeTimeDeltaArrayType, 'shape')
def overload_datetime_timedelta_arr_shape(A):
    return lambda A: (len(A._days_data),)


@overload_attribute(DatetimeTimeDeltaArrayType, 'nbytes')
def timedelta_arr_nbytes_overload(A):
    return (lambda A: A._days_data.nbytes + A._seconds_data.nbytes + A.
        _microseconds_data.nbytes + A._null_bitmap.nbytes)


def overload_datetime_timedelta_arr_sub(arg1, arg2):
    if (arg1 == datetime_timedelta_array_type and arg2 ==
        datetime_timedelta_type):

        def impl(arg1, arg2):
            bht__pss = arg1
            numba.parfors.parfor.init_prange()
            n = len(bht__pss)
            A = alloc_datetime_timedelta_array(n)
            for uvhh__kei in numba.parfors.parfor.internal_prange(n):
                A[uvhh__kei] = bht__pss[uvhh__kei] - arg2
            return A
        return impl


def create_cmp_op_overload_arr(op):

    def overload_date_arr_cmp(lhs, rhs):
        if op == operator.ne:
            onboy__aevz = True
        else:
            onboy__aevz = False
        if (lhs == datetime_timedelta_array_type and rhs ==
            datetime_timedelta_array_type):

            def impl(lhs, rhs):
                numba.parfors.parfor.init_prange()
                n = len(lhs)
                fhh__xkrd = bodo.libs.bool_arr_ext.alloc_bool_array(n)
                for uvhh__kei in numba.parfors.parfor.internal_prange(n):
                    hkjva__wcng = bodo.libs.array_kernels.isna(lhs, uvhh__kei)
                    tmtf__iabi = bodo.libs.array_kernels.isna(rhs, uvhh__kei)
                    if hkjva__wcng or tmtf__iabi:
                        doj__hqusb = onboy__aevz
                    else:
                        doj__hqusb = op(lhs[uvhh__kei], rhs[uvhh__kei])
                    fhh__xkrd[uvhh__kei] = doj__hqusb
                return fhh__xkrd
            return impl
        elif lhs == datetime_timedelta_array_type:

            def impl(lhs, rhs):
                numba.parfors.parfor.init_prange()
                n = len(lhs)
                fhh__xkrd = bodo.libs.bool_arr_ext.alloc_bool_array(n)
                for uvhh__kei in numba.parfors.parfor.internal_prange(n):
                    hwsy__seg = bodo.libs.array_kernels.isna(lhs, uvhh__kei)
                    if hwsy__seg:
                        doj__hqusb = onboy__aevz
                    else:
                        doj__hqusb = op(lhs[uvhh__kei], rhs)
                    fhh__xkrd[uvhh__kei] = doj__hqusb
                return fhh__xkrd
            return impl
        elif rhs == datetime_timedelta_array_type:

            def impl(lhs, rhs):
                numba.parfors.parfor.init_prange()
                n = len(rhs)
                fhh__xkrd = bodo.libs.bool_arr_ext.alloc_bool_array(n)
                for uvhh__kei in numba.parfors.parfor.internal_prange(n):
                    hwsy__seg = bodo.libs.array_kernels.isna(rhs, uvhh__kei)
                    if hwsy__seg:
                        doj__hqusb = onboy__aevz
                    else:
                        doj__hqusb = op(lhs, rhs[uvhh__kei])
                    fhh__xkrd[uvhh__kei] = doj__hqusb
                return fhh__xkrd
            return impl
    return overload_date_arr_cmp


timedelta_unsupported_attrs = ['asm8', 'resolution_string', 'freq',
    'is_populated']
timedelta_unsupported_methods = ['isoformat']


def _intstall_pd_timedelta_unsupported():
    from bodo.utils.typing import create_unsupported_overload
    for knny__lmupy in timedelta_unsupported_attrs:
        havj__mvqe = 'pandas.Timedelta.' + knny__lmupy
        overload_attribute(PDTimeDeltaType, knny__lmupy)(
            create_unsupported_overload(havj__mvqe))
    for yvxg__arirk in timedelta_unsupported_methods:
        havj__mvqe = 'pandas.Timedelta.' + yvxg__arirk
        overload_method(PDTimeDeltaType, yvxg__arirk)(
            create_unsupported_overload(havj__mvqe + '()'))


_intstall_pd_timedelta_unsupported()
